/* 
   Affix - Bluetooth Protocol Stack for Linux
   Copyright (C) 2001 Nokia Corporation
   Original Author: Dmitry Kasatkin <dmitry.kasatkin@nokia.com>

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 2 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program; if not, write to the Free Software Foundation, Inc.,
   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

/* 
   $Id: l2cap.h,v 1.12 2002/11/14 00:45:16 kds Exp $

   Link Layer Control and Adaptation Protocol

   Fixes:	Dmitry Kasatkin <dmitry.kasatkin@nokia.com>
*/

#ifndef L2CAP_H
#define L2CAP_H

#include <linux/types.h>
#include <linux/skbuff.h>
#include <linux/module.h>

#include <affix/bluetooth.h>
#include <affix/hci.h>


#define	L2CAP_HDR_LEN		4

//#define RTX_TIMEOUT		(30*HZ)
//#define ERTX_TIMEOUT		(30*HZ)
#define RTX_TIMEOUT		(60*HZ)
#define RTX_TIMEOUT_DISC	(2*HZ)
#define ERTX_TIMEOUT		(300*HZ)


/* Channel configuration states */
typedef volatile enum {
	CHN_CFG_START = 1,
	CHN_CFG_REQSENT,
	CHN_CFG_REQRECV,
	CHN_CFG_END
} l2cap_cfg_state;


/* Signalling channel commands' id's */

#define SIG_RESERVED		0x00
#define SIG_REJECT		0x01
#define SIG_CONREQ		0x02
#define SIG_CONRSP		0x03
#define SIG_CFGREQ		0x04
#define SIG_CFGRSP		0x05
#define SIG_DISCREQ		0x06
#define SIG_DISCRSP		0x07
#define SIG_ECHOREQ		0x08
#define SIG_ECHORSP		0x09
#define SIG_INFOREQ		0x0A
#define SIG_INFORSP		0x0B

/* some proprietary Affix commands */
#define SIG_SINGLEPING		0x80

/* Channels */

#define CID_NULL		0x0000
#define CID_SIGNAL		0x0001
#define CID_CONLESS		0x0002
#define CID_GROUP		CID_CONLESS
#define CID_BASE		0x0040

#define RES_SUCCESS		0x00


/* l2cap protocol header */
typedef struct {
	__u16	length;
	__u16	cid;
	__u8	data[0];
} __PACK__ l2cap_hdr_t;

/* Connection less header */
typedef struct {
	__u16	length;
	__u16	cid;
	__u16	psm;
	__u8	data[0];
} __PACK__ l2cap_grp_hdr_t;


typedef struct {
	__u8	code;
	__u8	id;
	__u16	length;
	__u8	data[0];
} __PACK__ l2cap_cmd_t;


typedef struct {
	__u8	code;
	__u8	id;
	__u16	length;
	__u16	reason;
#define	REJ_NOTSUP	0x0000
#define	REJ_MTUEXCEED	0x0001
#define	REJ_INVCID	0x0002
	__u8	data[0];
} __PACK__ l2cap_cmdrej_t;

typedef struct {
	__u8	code;
	__u8	id;
	__u16	length;
	__u16	psm;
	__u16	scid;
} __PACK__ l2cap_conreq_t;

typedef struct {
	__u8	code;
	__u8	id;
	__u16	length;
	__u16	dcid;
	__u16	scid;
	__u16	result;
#define CON_RSP_SUCCESS		0x0000
#define CON_RSP_PENDING		0x0001
#define CON_RSP_PSM		0x0002
#define CON_RSP_SECURITY	0x0003
#define CON_RSP_RESOURCE	0x0004
	__u16	status;
#define	CON_RSP_NOINFO		0x0000
#define	CON_RSP_AUTHEN		0x0001
#define	CON_RSP_AUTHOR		0x0002
} __PACK__ l2cap_conrsp_t;

typedef struct {
	__u8	code;
	__u8	id;
	__u16	length;
	__u16	dcid;
	__u16	flags;
#define CFG_REQ_MORE		0x0001
	__u8	options[0];
} __PACK__ l2cap_cfgreq_t;

typedef struct {
	__u8	code;
	__u8	id;
	__u16	length;
	__u16	scid;
	__u16	flags;
	__u16	result;
#define	CFG_RSP_SUCCESS		0x0000
#define CFG_RSP_PARAMETERS	0x0001
#define CFG_RSP_REASON		0x0002
#define CFG_RSP_OPTIONS		0x0003
	__u8	options[0];
} __PACK__ l2cap_cfgrsp_t;

typedef struct {
	__u8	code;
	__u8	id;
	__u16	length;
	__u16	dcid;
	__u16	scid;
} __PACK__ l2cap_discreq_t;

typedef struct {
	__u8	code;
	__u8	id;
	__u16	length;
	__u16	dcid;
	__u16	scid;
} __PACK__ l2cap_discrsp_t;

typedef struct {
	__u8	code;
	__u8	id;
	__u16	length;
	__u16	infotype;
#define INFO_REQ_MTU	0x0001
} __PACK__ l2cap_inforeq_t;

typedef struct {
	__u8	code;
	__u8	id;
	__u16	length;
	__u16	infotype;
	__u16	result;
#define INFO_RSP_SUCCESS	0x0000
#define INFO_RSP_NOTSUPPORTED	0x0001
	__u8	data[0];
/* here is MTU - 2 bytes */
} __PACK__ l2cap_inforsp_t;


/* Configuration parameter options */
typedef struct {
	__u8	type;
#define CFG_OPT_MTU		0x01
#define CFG_OPT_FLUSHTO		0x02
#define CFG_OPT_QOS		0x03
#define	CFG_OPT_SKIP		0x80
	__u8	length;
	__u8	data[0];
} __PACK__ l2cap_cfgopt_t;

/* QoS options */
typedef struct {
	__u8	flags;
	__u8	service_type;
#define QOS_NO_TRAFFIC		0x00
#define QOS_BEST_EFFORT		0x01
#define QOS_GUARANTEED		0x02
	__u32	token_rate;
	__u32	token_size;
	__u32	bandwidth;
	__u32	latency;
	__u32	delay_variation;
} __PACK__ l2cap_qos_t;


/**************************************************************************************/

typedef struct {
	__u32		flags;	/* what should be negotiated */
	__u16		mtu;
	__u16		flush_to;
	l2cap_qos_t	qos;
	__u16		link_to;
} l2cap_cfg_opt;


typedef struct l2cap_ch  l2cap_ch;

typedef struct {
	struct	module	*owner;
	int (*data_ind)(l2cap_ch *ch, struct sk_buff *skb);
	int (*connect_ind)(l2cap_ch *ch);
	int (*connect_cfm)(l2cap_ch *ch, __u16 result, __u16 status);
	int (*config_ind)(l2cap_ch *ch);
	int (*config_cfm)(l2cap_ch *ch, __u16 result);
	int (*disconnect_ind)(l2cap_ch *ch);
} l2cap_proto_ops;

typedef struct {
	struct list_head	q;
	__u16			psm;
	l2cap_proto_ops		*ops;
} l2cap_proto;


/*
   l2cap_ch describe the L2CAP channel
*/
struct l2cap_ch{
	struct list_head	q;		/* for queueing */
	
	atomic_t		refcnt;		/* number of users */
	con_state		state;		/* State of the channel	*/

	void			*priv;		/* Private field for upper layer */

	/* Upper protocol stuff */
	BD_ADDR			bda;
	__u16			psm;
	l2cap_proto_ops		*ops;

	/* L2CAP Protocol stuff */
	__u16			lcid;		/* local CID		*/
	__u16			rcid;		/* remote CID		*/

	/* request management */
	struct timer_list	timer;
	__u8			reqid;		/* commands request id	*/
	__u8			rspid;		/* commands response id	*/

	/* Configuration stuff */
	l2cap_cfg_state		out_state;	/* Request path	*/
	l2cap_cfg_opt		out_flow;	/* to remote device */

	l2cap_cfg_state		in_state;	/* Response path */
	l2cap_cfg_opt		in_flow;	/* from remote device */

	unsigned int		flags;		/* configuration flags 	*/
#define L2CAP_SERVER		0
#define L2CAP_CONFIG		1

	/* HCI related */
	hci_struct		*hci;		/* device handle */
	hci_con			*con;		/* contains BD_ADDR */
	__u8			hdrlen;
};



/* L2CAP definition */
#define PSM_SDP		0x01
#define PSM_RFCOMM	0x03

/*Internal */

extern __u8	cmd_id;


/* protocol management */
l2cap_proto * l2cap_proto_lookup(__u16 psm);
int l2ca_register_protocol(__u16 psm, l2cap_proto_ops *pops);
void l2ca_unregister_protocol(__u16 psm);


/* channel management */
l2cap_ch *l2ca_create(__u16 psm, l2cap_proto_ops *pops);
void l2cap_destroy(l2cap_ch *ch);
void __l2cap_destroy(l2cap_ch *ch);
void l2ca_put(l2cap_ch *ch);

void def_cfg_opt(l2cap_cfg_opt *opt);
int decode_cfg_opt(struct sk_buff *skb, l2cap_cfg_opt *opt);
int encode_cfg_opt(__u8 *options, l2cap_cfg_opt *opt);

/* For upper layer */
int l2ca_connect_req(l2cap_ch *ch, BD_ADDR *bdaddr);
int l2ca_connect_rsp(l2cap_ch *ch, __u16 response, __u16 status);
int l2ca_config_req(l2cap_ch *ch);
int l2ca_config_rsp(l2cap_ch *ch, __u16 result);
int l2ca_disconnect_req(l2cap_ch *ch);
//int l2ca_disconnect_rsp(l2cap_ch *ch);
int l2ca_send_data(l2cap_ch *ch, struct sk_buff *skb);
int l2ca_ping(l2cap_ch *ch, struct sk_buff *skb);
int l2ca_singleping(l2cap_ch *ch, struct sk_buff *skb);

/* to upper layer */
int l2ca_data_ind(l2cap_ch *ch, struct sk_buff *skb);
int l2ca_connect_ind(l2cap_ch *ch);
int l2ca_connect_cfm(l2cap_ch *ch, __u16 result, __u16 status);
int l2ca_config_ind(l2cap_ch *ch);
int l2ca_config_cfm(l2cap_ch *ch, __u16 result);
int l2ca_disconnect_ind(l2cap_ch *ch);

/* Signalling */
int l2cap_connect_req(hci_con *con, __u8 id, __u16 psm, __u16 scid);
int l2cap_connect_rsp(hci_con *con, __u8 id, __u16 scid, __u16 dcid, __u16 result, __u16 status);
int l2cap_command_rej(hci_con *con, __u8 id, __u16 reason, __u16 length, __u8 *data);
int l2cap_config_req(hci_con *con, __u8 id, __u16 dcid, __u16 flags, __u8 olen, __u8 *options);
int l2cap_config_rsp(hci_con *con, __u8 id, __u16 dcid, __u16 flags, __u16 result, __u8 olen, __u8 *options);
int l2cap_disconnect_req(hci_con *con, __u8 id, __u16 dcid, __u16 scid);
int l2cap_disconnect_rsp(hci_con *con, __u8 id, __u16 scid, __u16 dcid);
int l2cap_echo_req(hci_con *con, __u8 id, struct sk_buff *skb, __u8 code);
int l2cap_echo_rsp(hci_con *con, __u8 id, __u16 length, __u8 *data);
int l2cap_info_req(hci_con *con, __u8 id, __u16 infotype);
int l2cap_info_rsp(hci_con *con, __u8 id, __u16 infotype, __u16 length, __u8 *data);


/*
  inline functions
*/
static inline void l2ca_hold(l2cap_ch *ch)
{
	__MOD_INC_USE_COUNT(ch->ops->owner);
	atomic_inc(&ch->refcnt);
}



/*
 * l2cap timers
 */
#define L2CAP_TIMEOUT_CONRSP	0x01
#define L2CAP_TIMEOUT_CFGRSP	0x02
#define L2CAP_TIMEOUT_DISCRSP	0x03

static inline void l2cap_start_timer(l2cap_ch *ch, unsigned long len)
{
	if (!del_timer(&ch->timer))
		l2ca_hold(ch);
	mod_timer(&ch->timer, jiffies+len);
}

static inline void l2cap_stop_timer(l2cap_ch *ch)
{
	if (del_timer(&ch->timer))
		l2ca_put(ch);
}


static inline void l2ca_setmtu(l2cap_ch *ch, int mtu)
{
	ch->out_flow.mtu = mtu;
	set_bit(CFG_OPT_MTU, &ch->out_flow.flags);
}

static inline int l2ca_getmtu(l2cap_ch *ch)
{
	return ch->in_flow.mtu;
}

static inline void l2ca_setqos(l2cap_ch *ch, l2cap_qos_t *qos)
{
	ch->out_flow.qos = *qos;
	set_bit(CFG_OPT_QOS, &ch->out_flow.flags);
}

static inline void l2cap_config_start(l2cap_ch *ch)
{
	ENTERSTATE(ch, CON_CONFIG);
	ch->out_state = CHN_CFG_START;
	ch->in_state = CHN_CFG_START;
}

static inline int l2ca_server(l2cap_ch *ch)
{
	return test_bit(L2CAP_SERVER, &ch->flags);
}

static inline __u8 get_id(l2cap_ch *ch)
{
	ch->reqid = cmd_id++;
	if (cmd_id == 0)
		cmd_id++;

	return ch->reqid;
}

#endif
